@@ -1,71 +1,47 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Creator.
-**
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #pragma once
 
 #include "gitsettings.h"
+#include "git_global.h"
 #include "commitdata.h"
 
 #include <coreplugin/editormanager/ieditor.h>
 #include <coreplugin/iversioncontrol.h>
+
+#include <texteditor/texteditorconstants.h>
+
 #include <vcsbase/vcsbaseclient.h>
 
-#include <utils/fileutils.h>
-#include <utils/futuresynchronizer.h>
-
-#include <QObject>
-#include <QString>
 #include <QStringList>
 #include <QWidget>
 
 QT_BEGIN_NAMESPACE
-class QProcessEnvironment;
 class QMenu;
 QT_END_NAMESPACE
 
-namespace Core { class ICore; }
+namespace Tasking { class GroupItem; }
 
-namespace VcsBase {
-class VcsCommand;
-class SubmitFileModel;
-class VcsBaseEditorWidget;
-}
+namespace VcsBase { class SubmitFileModel; }
 
-namespace DiffEditor {
-class ChunkSelection;
-class DiffEditorController;
-}
-
-namespace Git {
-namespace Internal {
+namespace Git::Internal {
 
 class CommitData;
 class GitBaseDiffEditorController;
 class GitSubmitEditorPanelData;
 class Stash;
 
+struct ColorNames
+{
+    QString author;
+    QString date;
+    QString hash;
+    QString decoration;
+    QString subject;
+    QString body;
+};
+
 enum StatusMode
 {
     ShowAll = 0,
@@ -79,12 +55,6 @@ enum StashFlag {
     NoPrompt       = 0x02
 };
 
-enum PushFailure {
-    Unknown,
-    NonFastForward,
-    NoRemoteBranch
-};
-
 class SubmoduleData
 {
 public:
@@ -101,17 +71,27 @@ public:
     UpstreamStatus() = default;
     UpstreamStatus(int ahead, int behind) : ahead(ahead), behind(behind) {}
 
-    int ahead = 0;
-    int behind = 0;
+    int ahead = -1;
+    int behind = -1;
 };
 
-class GitClient : public VcsBase::VcsBaseClientImpl
-{
-    Q_OBJECT
+struct Author {
+    bool operator==(const Author &other) const {
+        return name == other.name && email == other.email;
+    }
+    bool operator!=(const Author &other) const {
+        return !operator==(other);
+    }
+    QString name;
+    QString email;
+};
 
+class GITSHARED_EXPORT GitClient : public VcsBase::VcsBaseClientImpl
+{
 public:
     enum CommandInProgress { NoCommand, Revert, CherryPick,
                              Rebase, Merge, RebaseMerge };
+    enum GitKLaunchTrial { Bin, ParentOfBin, SystemPath, None };
 
     class StashInfo
     {
@@ -124,8 +104,6 @@ public:
                   StashFlag flag = Default, PushAction pushAction = NoPush);
         bool stashingFailed() const;
         void end();
-        StashResult result() const { return m_stashResult; }
-        QString stashMessage() const { return m_message; }
 
     private:
         void stashPrompt(const QString &command, const QString &statusOutput, QString *errorMessage);
@@ -138,54 +116,70 @@ public:
         PushAction m_pushAction = NoPush;
     };
 
-    explicit GitClient(GitSettings *settings);
-    static GitClient *instance();
-    static GitSettings &settings();
+    struct ModificationInfo
+    {
+        Utils::FilePath rootPath;
+        QHash<QString, Core::IVersionControl::FileState> modifiedFiles;
+    };
 
-    Utils::FilePath vcsBinary() const override;
-    unsigned gitVersion(QString *errorMessage = nullptr) const;
+    GitClient();
+    ~GitClient();
 
-    VcsBase::VcsCommand *vcsExecAbortable(const Utils::FilePath &workingDirectory,
-                                          const QStringList &arguments,
-                                          bool isRebase = false,
-                                          QString abortCommand = {});
+    Utils::FilePath vcsBinary(const Utils::FilePath &forDirectory) const override;
+
+    void vcsExecAbortable(const Utils::FilePath &workingDirectory, const QStringList &arguments,
+                          bool isRebase = false, const QString &abortCommand = {},
+                          const QObject *context = nullptr,
+                          const VcsBase::CommandHandler &handler = {});
 
     Utils::FilePath findRepositoryForDirectory(const Utils::FilePath &directory) const;
-    QString findGitDirForRepository(const Utils::FilePath &repositoryDir) const;
+    Utils::FilePath findGitDirForRepository(const Utils::FilePath &repositoryDir) const;
     bool managesFile(const Utils::FilePath &workingDirectory, const QString &fileName) const;
     Utils::FilePaths unmanagedFiles(const Utils::FilePaths &filePaths) const;
+    Core::IVersionControl::FileState modificationState(const Utils::FilePath &workingDirectory,
+                         const Utils::FilePath &fileName) const;
+    void monitorDirectory(const Utils::FilePath &path);
+    void stopMonitoring(const Utils::FilePath &path);
 
-    void diffFile(const Utils::FilePath &workingDirectory, const QString &fileName) const;
+    enum DiffMode { Unstaged, Staged };
+    void diffFile(const Utils::FilePath &workingDirectory, const QString &fileName,
+                  DiffMode diffMode = Unstaged) const;
     void diffFiles(const Utils::FilePath &workingDirectory,
                    const QStringList &unstagedFileNames,
                    const QStringList &stagedFileNames) const;
     void diffProject(const Utils::FilePath &workingDirectory,
-                     const QString &projectDirectory) const;
-    void diffRepository(const Utils::FilePath &workingDirectory) const
+                     const QString &projectDirectory, DiffMode diffMode = Unstaged) const;
+    void diffUnstagedRepository(const Utils::FilePath &workingDirectory) const
     {
-        return diffRepository(workingDirectory, {}, {});
+        return diffRepository(workingDirectory, {}, {}, Unstaged);
+    }
+    void diffStagedRepository(const Utils::FilePath &workingDirectory) const
+    {
+        return diffRepository(workingDirectory, {}, {}, Staged);
     }
     void diffRepository(const Utils::FilePath &workingDirectory,
                         const QString &leftCommit,
-                        const QString &rightCommit) const;
+                        const QString &rightCommit,
+                        DiffMode diffMode = Unstaged) const;
     void diffBranch(const Utils::FilePath &workingDirectory,
                     const QString &branchName) const;
     void merge(const Utils::FilePath &workingDirectory, const QStringList &unmergedFileNames = {});
 
     void status(const Utils::FilePath &workingDirectory) const;
+    void fullStatus(const Utils::FilePath &workingDirectory) const;
     void log(const Utils::FilePath &workingDirectory, const QString &fileName = {},
              bool enableAnnotationContextMenu = false, const QStringList &args = {});
     void reflog(const Utils::FilePath &workingDirectory, const QString &branch = {});
-    VcsBase::VcsBaseEditorWidget *annotate(const Utils::FilePath &workingDir, const QString &file,
-                                           const QString &revision = {}, int lineNumber = -1,
-                                           const QStringList &extraOptions = {}) override;
+    void annotate(const Utils::FilePath &workingDir, const QString &file,
+                  int lineNumber = -1, const QString &revision = {},
+                  const QStringList &extraOptions = {}, int firstLine = -1) override;
     void reset(const Utils::FilePath &workingDirectory, const QString &argument, const QString &commit = {});
     void removeStaleRemoteBranches(const Utils::FilePath &workingDirectory, const QString &remote);
     void recoverDeletedFiles(const Utils::FilePath &workingDirectory);
     void addFile(const Utils::FilePath &workingDirectory, const QString &fileName);
-    bool synchronousLog(const Utils::FilePath &workingDirectory, const QStringList &arguments,
-                        QString *output, QString *errorMessage = nullptr,
-                        unsigned flags = 0);
+    Utils::Result<QString> synchronousLog(const Utils::FilePath &workingDirectory,
+                                          const QStringList &arguments,
+                                          VcsBase::RunFlags flags = VcsBase::RunFlags::None);
     bool synchronousAdd(const Utils::FilePath &workingDirectory, const QStringList &files,
                         const QStringList &extraOptions = {});
     bool synchronousDelete(const Utils::FilePath &workingDirectory,
@@ -199,14 +193,16 @@ public:
     bool synchronousCleanList(const Utils::FilePath &workingDirectory, const QString &modulePath,
                               QStringList *files, QStringList *ignoredFiles, QString *errorMessage);
     bool synchronousApplyPatch(const Utils::FilePath &workingDirectory, const QString &file,
-                               QString *errorMessage, const QStringList &extraArguments = {});
+                               QString *errorMessage, const QStringList &extraArguments = {}) const;
     bool synchronousInit(const Utils::FilePath &workingDirectory);
+    bool synchronousAddGitignore(const Utils::FilePath &workingDirectory);
     bool synchronousCheckoutFiles(const Utils::FilePath &workingDirectory, QStringList files = {},
                                   QString revision = {}, QString *errorMessage = nullptr,
                                   bool revertStaging = true);
     enum class StashMode { NoStash, TryStash };
-    VcsBase::VcsCommand *checkout(const Utils::FilePath &workingDirectory, const QString &ref,
-                                  StashMode stashMode = StashMode::TryStash);
+    void checkout(const Utils::FilePath &workingDirectory, const QString &ref,
+                  StashMode stashMode = StashMode::TryStash, const QObject *context = nullptr,
+                  const VcsBase::CommandHandler &handler = {});
 
     QStringList setupCheckoutArguments(const Utils::FilePath &workingDirectory, const QString &ref);
     void updateSubmodulesIfNeeded(const Utils::FilePath &workingDirectory, bool prompt);
@@ -234,7 +230,6 @@ public:
                            QString *output, QString *errorMessage) const;
     bool synchronousForEachRefCmd(const Utils::FilePath &workingDirectory, QStringList args,
                                QString *output, QString *errorMessage = nullptr) const;
-    VcsBase::VcsCommand *asyncForEachRefCmd(const Utils::FilePath &workingDirectory, QStringList args) const;
     bool synchronousRemoteCmd(const Utils::FilePath &workingDirectory, QStringList remoteArgs,
                               QString *output = nullptr, QString *errorMessage = nullptr,
                               bool silent = false) const;
@@ -245,7 +240,7 @@ public:
                                            QString *errorMessage = nullptr) const;
     SubmoduleDataMap submoduleList(const Utils::FilePath &workingDirectory) const;
     QByteArray synchronousShow(const Utils::FilePath &workingDirectory, const QString &id,
-                               unsigned flags = 0) const;
+                               VcsBase::RunFlags flags = VcsBase::RunFlags::None) const;
 
     bool synchronousRevListCmd(const Utils::FilePath &workingDirectory, const QStringList &extraArguments,
                                QString *output, QString *errorMessage = nullptr) const;
@@ -265,11 +260,9 @@ public:
     QString synchronousTopic(const Utils::FilePath &workingDirectory) const;
     bool synchronousRevParseCmd(const Utils::FilePath &workingDirectory, const QString &ref,
                                 QString *output, QString *errorMessage = nullptr) const;
-    QString synchronousTopRevision(const Utils::FilePath &workingDirectory, QDateTime *dateTime = nullptr);
-    void synchronousTagsForCommit(const Utils::FilePath &workingDirectory, const QString &revision,
-                                  QString &precedes, QString &follows) const;
+    Tasking::GroupItem topRevision(const Utils::FilePath &workingDirectory,
+        const std::function<void(const QString &, const QDateTime &)> &callback);
     bool isRemoteCommit(const Utils::FilePath &workingDirectory, const QString &commit);
-    bool isFastForwardMerge(const Utils::FilePath &workingDirectory, const QString &branch);
 
     void fetch(const Utils::FilePath &workingDirectory, const QString &remote);
     void pull(const Utils::FilePath &workingDirectory, bool rebase);
@@ -297,30 +290,26 @@ public:
     void subversionDeltaCommit(const Utils::FilePath &workingDirectory) const;
 
     void stashPop(const Utils::FilePath &workingDirectory, const QString &stash = {});
-    void revert(const QStringList &files, bool revertStaging);
-    bool synchronousStashList(const Utils::FilePath &workingDirectory, QList<Stash> *stashes,
-                              QString *errorMessage = nullptr) const;
-    // Resolve a stash name from message (for IVersionControl's names).
-    bool stashNameFromMessage(const Utils::FilePath &workingDirectory, const QString &messge, QString *name,
-                              QString *errorMessage = nullptr) const;
+    void revertFiles(const QStringList &files, bool revertStaging);
+    QList<Stash> synchronousStashList(const Utils::FilePath &workingDirectory) const;
 
     QString readGitVar(const Utils::FilePath &workingDirectory, const QString &configVar) const;
     QString readConfigValue(const Utils::FilePath &workingDirectory, const QString &configVar) const;
+    QChar commentChar(const Utils::FilePath &workingDirectory);
     void setConfigValue(const Utils::FilePath &workingDirectory, const QString &configVar,
                         const QString &value) const;
 
-    QTextCodec *encoding(const Utils::FilePath &workingDirectory, const QString &configVar) const;
-    bool readDataFromCommit(const Utils::FilePath &repoDirectory, const QString &commit,
-                            CommitData &commitData, QString *errorMessage = nullptr,
-                            QString *commitTemplate = nullptr);
-    bool getCommitData(const Utils::FilePath &workingDirectory, QString *commitTemplate,
-                       CommitData &commitData, QString *errorMessage);
+    Utils::Result<CommitData> enrichCommitData(const Utils::FilePath &repoDirectory,
+                                               const QString &commit,
+                                               const CommitData &commitDataIn);
+    Utils::Result<CommitData> getCommitData(CommitType commitType,
+                                            const Utils::FilePath &workingDirectory);
 
     bool addAndCommit(const Utils::FilePath &workingDirectory,
                       const GitSubmitEditorPanelData &data,
                       CommitType commitType,
-                      const QString &amendSHA1,
-                      const QString &messageFile,
+                      const QString &amendHash,
+                      const Utils::FilePath &messageFile,
                       VcsBase::SubmitFileModel *model);
 
     enum StatusResult { StatusChanged, StatusUnchanged, StatusFailed };
@@ -332,8 +321,6 @@ public:
 
     void continueCommandIfNeeded(const Utils::FilePath &workingDirectory, bool allowContinue = true);
 
-    QString extendedShowDescription(const Utils::FilePath &workingDirectory, const QString &text) const;
-
     void launchGitK(const Utils::FilePath &workingDirectory, const QString &fileName) const;
     void launchGitK(const Utils::FilePath &workingDirectory) const { launchGitK(workingDirectory, QString()); }
     bool launchGitGui(const Utils::FilePath &workingDirectory);
@@ -345,53 +332,52 @@ public:
     QStringList synchronousRepositoryBranches(const QString &repositoryURL,
                                               const Utils::FilePath &workingDirectory = {}) const;
 
-    Utils::Environment processEnvironment() const override;
+    Utils::Environment processEnvironment(const Utils::FilePath &appliedTo) const override;
 
     bool beginStashScope(const Utils::FilePath &workingDirectory, const QString &command,
                          StashFlag flag = Default, PushAction pushAction = NoPush);
-    StashInfo &stashInfo(const Utils::FilePath &workingDirectory);
     void endStashScope(const Utils::FilePath &workingDirectory);
     bool isValidRevision(const QString &revision) const;
     void handleMergeConflicts(const Utils::FilePath &workingDir, const QString &commit,
                               const QStringList &files, const QString &abortCommand);
-    void addFuture(const QFuture<void> &future);
 
     static QString msgNoChangedFiles();
     static QString msgNoCommits(bool includeRemote);
-    void show(const QString &source, const QString &id, const QString &name = {});
+    void show(const Utils::FilePath &source, const QString &id, const QString &name = {});
     void archive(const Utils::FilePath &workingDirectory, QString commit);
 
-    VcsBase::VcsCommand *asyncUpstreamStatus(const Utils::FilePath &workingDirectory,
-                                             const QString &branch, const QString &upstream);
-
     enum class BranchTargetType { Remote, Commit };
     static QString suggestedLocalBranchName(
             const Utils::FilePath &workingDirectory, const QStringList &existingLocalNames,
             const QString &target, BranchTargetType targetType);
-    static void addChangeActions(QMenu *menu, const QString &source, const QString &change);
-    static Utils::FilePath fileWorkingDirectory(const QString &file);
+    static void addChangeActions(QMenu *menu, const Utils::FilePath &source, const QString &change);
+    static Utils::FilePath fileWorkingDirectory(const Utils::FilePath &file);
     enum class ShowEditor { OnlyIfDifferent, Always };
     Core::IEditor *openShowEditor(const Utils::FilePath &workingDirectory, const QString &ref,
-                                  const QString &path, ShowEditor showSetting = ShowEditor::Always);
+                                  const Utils::FilePath &path, ShowEditor showSetting = ShowEditor::Always);
+
+    Author parseAuthor(const QString &authorInfo);
+    Author getAuthor(const Utils::FilePath &workingDirectory);
+
+    QTextCodec *defaultCommitEncoding() const;
+    enum EncodingType { EncodingSource, EncodingLogOutput, EncodingCommit, EncodingDefault };
+    QTextCodec *encoding(EncodingType encodingType, const Utils::FilePath &source = {}) const;
+
+    void readConfigAsync(const Utils::FilePath &workingDirectory, const QStringList &arguments,
+                         const VcsBase::CommandHandler &handler) const;
+
+    static QString styleColorName(TextEditor::TextStyle style);
+    static ColorNames colorNames();
 
 private:
+    static GitSettings &settings();
+
     void finishSubmoduleUpdate();
-    void chunkActionsRequested(QMenu *menu, int fileIndex, int chunkIndex,
-                               const DiffEditor::ChunkSelection &selection);
 
-    void stage(DiffEditor::DiffEditorController *diffController,
-               const QString &patch, bool revert);
-
-    enum CodecType { CodecSource, CodecLogOutput, CodecNone };
-    QTextCodec *codecFor(CodecType codecType, const Utils::FilePath &source = {}) const;
-
-    void requestReload(const QString &documentId, const QString &source, const QString &title,
-                       const Utils::FilePath &workingDirectory,
+    void requestReload(const QString &documentId, const Utils::FilePath &source,
+                       const QString &title, const Utils::FilePath &workingDirectory,
                        std::function<GitBaseDiffEditorController *(Core::IDocument *)> factory) const;
 
-    // determine version as '(major << 16) + (minor << 8) + patch' or 0.
-    unsigned synchronousGitVersion(QString *errorMessage = nullptr) const;
-
     QString readOneLine(const Utils::FilePath &workingDirectory, const QStringList &arguments) const;
 
     enum RevertResult { RevertOk, RevertUnchanged, RevertCanceled, RevertFailed };
@@ -399,15 +385,20 @@ private:
                          bool *isDirectory,
                          QString *errorMessage,
                          bool revertStaging);
-    void connectRepositoryChanged(const QString & repository, VcsBase::VcsCommand *cmd);
     bool executeAndHandleConflicts(const Utils::FilePath &workingDirectory, const QStringList &arguments,
                                    const QString &abortCommand = {}) const;
-    bool tryLauchingGitK(const Utils::Environment &env,
-                         const Utils::FilePath &workingDirectory,
-                         const QString &fileName,
-                         const Utils::FilePath &gitBinDirectory) const;
+    void tryLaunchingGitK(const Utils::Environment &env,
+                          const Utils::FilePath &workingDirectory,
+                          const QString &fileName,
+                          GitKLaunchTrial trial = GitKLaunchTrial::Bin) const;
+    void handleGitKFailedToStart(const Utils::Environment &env,
+                                 const Utils::FilePath &workingDirectory,
+                                 const QString &fileName,
+                                 const GitKLaunchTrial oldTrial,
+                                 const Utils::FilePath &oldGitBinDir) const;
     bool cleanList(const Utils::FilePath &workingDirectory, const QString &modulePath,
                    const QString &flag, QStringList *files, QString *errorMessage);
+    void updateModificationInfos();
 
     enum ContinueCommandMode {
         ContinueOnly,
@@ -419,23 +410,24 @@ private:
                                     QString msgBoxText, const QString &buttonName,
                                     const QString &gitCommand, ContinueCommandMode continueMode);
 
-    mutable Utils::FilePath m_gitVersionForBinary;
-    mutable unsigned m_cachedGitVersion = 0;
+    void setupTimer();
+    mutable QMap<Utils::FilePath, Utils::FilePath> m_gitExecutableCache;
 
     QString m_gitQtcEditor;
     QMap<Utils::FilePath, StashInfo> m_stashInfo;
-    QString m_pushFallbackCommand;
+    QHash<Utils::FilePath, ModificationInfo> m_modifInfos;
+    std::unique_ptr<QTimer> m_timer;
     QString m_diffCommit;
     Utils::FilePaths m_updatedSubmodules;
     bool m_disableEditor = false;
-    Utils::FutureSynchronizer m_synchronizer; // for commit updates
 };
 
+GITSHARED_EXPORT GitClient &gitClient();
+
 class GitRemote : public Core::IVersionControl::RepoUrl
 {
 public:
     GitRemote(const QString &location);
 };
 
-} // namespace Internal
-} // namespace Git
+} // Git::Internal
